"""
DtmModelInterface is an interface to create new DTM modules, these modules
should train the model and return values to front end.
"""

class DtmModelInterface:
    """
    DtmModelInterface defines methods and attributes that a module should have \
    in order to be passed to front end.

    :param corpus: Each item from the list is one document from corpus.
    :type corpus: list[str]

    :param dates: List of timestamps for each document in corpus, each \
    date's position should match with its respective text.
    :type dates: list[str]

    :param date_format: The date format used in `dates`.
    :type date_format: str

    :param freq: The frequency used to group texts.
    :type freq: str

    :param n_topics: Number of topics that the DTM model should find. The \
    default value is 100.
    :type n_topics: int, optional

    :param sep: Separator used to split each word, the default value is any \
    blank space.
    :type sep: str, optional

    :param workers: Number of workers (cpus) to use. If not provided, it \
    will use the value of multiprocessing.cpu_count()
    :type workers: int, optional

    """

    @property
    def n_timeslices(self):
        """
        This attribute should return the number of timeslices found.

        :return: It should return the number of time slices found in corpus.
        :rtype: int

        """
        pass


    def get_results(self):
        """
        This method should return a table representing the evolution of each \
        topic over time.

        :return: It must return a Pandas' dataframe where rows represents \
        different time slices and they are sorted by date, it must have one \
        column `data` and the remaining columns numbered from 0 to k (number \
        of topics - 1) that holds weights of each topic in that period.
        :rtype: pandas.core.frame.DataFrame

        """
        pass


    def get_topic_words(self, topic_id, i, n):
        """
        This method should return the top n words that better describes the \
        topic in a specific time slice.

        :param topic_id: The id of the desired topic.
        :type topic_id: int

        :param i: The position of the desired timeslice in chronological \
        order the first (oldest) time slice is indexed by 1.
        :type i: int

        :param n: This specifies how many words that better describes the \
        topic at a specific time slice should be returned.
        :type n: int

        :return: It returns a list of top n words that best describes the \
        requested topic in a specific time.
        :rtype: list[str]

        """
        pass


    def prepare_args(self, i):
        """
        This method should return a dictionary with all necessary values to \
        call PyLdaVis.prepare method.

        :param i: The position of the desired timeslice in chronological \
        order, the first (oldest) time slice is indexed by 1.
        :type i: int

        :return: It returns a dictionary ready to be passed to PyLdaVis
        :rtype: dict[str, any]

        """
        pass


    def train(self):
        """
        This method should train the dtm model.

        :return: nothing
        :rtype: None

        """
        pass



"""
LdaSeqModel brings the Gensim's LdaSeqModel functionalities to our library.

Its main advantage over other models is that it can detect changes in the \
vocabulary used to describe each topic over time, making it more precise in \
classifying each topic. But it is slower to run.
"""
# IMPORTS
from gensim.corpora.dictionary import Dictionary
from gensim.models.ldaseqmodel import LdaSeqModel as LdaSeqModel_
from lda_over_time.models.dtm_model_interface import DtmModelInterface
from multiprocessing import cpu_count

import numpy as np
import pandas as pd
import warnings


# TYPING
from typing import List, Optional


warnings.filterwarnings("ignore")


class LdaSeqModel(DtmModelInterface):
    """
    LdaSeqModel is a model that uses the Gensim's LdaSeqModel, which \
    supports the variance along time of the way that a certain topic is \
    approached (it can detect better the change of vocabulary to speak a \
    certain topic).

    With this feature, it may be more precise than PrevalenceModel, but \
    it is slower.

    :param corpus: List of documents' texts.
    :type corpus: list[str]

    :param dates: List of documents' publishing dates.
    :type dates: list[str]

    :param date_format: The date format used in `dates`, e.g. "%Y/%m/%d" for \
    "YYYY/MM/DD" format. More info at `documentation`_.
    :type date_format: str

    :param freq: The frequency used to group texts, e.g. "1M15D" for a \
    frequency of a month and 15 days. Useful notations: \
    day = "D"\
    month = "M"; \
    year = "Y". \
    More info at `pandas`_
    :type freq: str

    :param n_topics: Number of topics that the DTM model should find. The \
    default value is 100.
    :type n_topics: int, optional

    :param sep: Separator used to split each word, the default value is any \
    blank space.
    :type sep: str, optional

    :param workers: Number of workers (cpus) to use. If not provided, it will \
    use the total number of threads on running machine.
    :type workers: int, optional

    :return: Nothing
    :rtype: None

    .. _documentation: https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
    .. _pandas: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
    """

    def __init__(self,
                 corpus: List[str],
                 dates: List[str],
                 date_format: str,
                 freq: str,
                 n_topics: int = 100,
                 sep: Optional[str] = None,
                 workers: Optional[int] = None):
        """
        I initialize the variables to train model.
        """
        # Save arguments inside object
        self.corpus = corpus
        self.dates = dates
        self.date_format = date_format
        self.freq = freq
        self.n_topics = n_topics
        self.sep = sep

        # Assign workers equals to arguments' if passed or set default
        self.workers = workers if workers is not None else cpu_count()


    def __prepare_corpus(self, corpus, dates, date_format, freq, sep):
        """
        This method prepares the corpus and dates in a supported format to
        LdaSeqModel.
        """
        # load dates with pandas with given format
        parsed_dates = pd.to_datetime(dates, format=date_format)

        # sort documents by date
        date_doc_df = pd.DataFrame({'date': parsed_dates,
                                    'text': corpus}).sort_values('date')

        # group by frequency
        grouped = date_doc_df.groupby(pd.Grouper(key='date', freq=freq),
                                      sort=True)

        # get length of each timeslice ordered by time
        timeslices = list(grouped.size().values)

        # get timestamps for all time slices
        timestamps = list(grouped.groups.keys())

        # Split words for each document
        data = [doc.split(sep) for doc in date_doc_df['text'].values]

        # Train Dictionary
        dictionary = Dictionary(data)

        # Convert corpus to list of BoWs
        corpus_bow = [dictionary.doc2bow(doc) for doc in data]

        # return values
        return corpus_bow, dictionary, timestamps, timeslices


    def __prepare_results(self,
                          model,
                          timestamps,
                          n_documents,
                          n_topics,
                          time_slices):
        """
        This method extracts the main topic for each document in corpus and \
        calculates the proportion of each topic in a specific time.
        """
        #
        # TODO: find better way of doing this (more performatic)
        #
        # get main topic for each document
        main_topics = []
        for i in range(n_documents):

            # get the distribution of topics over document
            distribution = model.doc_topics(i)

            # get main topic
            main = distribution.argmax()

            # save main topic
            main_topics.append(main)

        # calculate initial and final position to calculate for each time slice
        accumulated = np.add.accumulate([0] + time_slices)
        start_end = zip(accumulated[:-1], accumulated[1:])

        # list of proportions for each time slice
        proportions = []

        # calculate proportion for each timeslice
        for ini, end in start_end:

            # count occurrences for each topic
            count = np.bincount(main_topics[ini:end])

            # fill missing topics with zero
            count = np.append(count, [0] * (n_topics - len(count)))

            # calculate proportion of each topic in a specific time slice
            proportion = count / np.sum(count)

            # append a dictionary of dict[topic, proportion]
            proportions.append(dict(zip(range(n_topics), proportion)))

        # create dataframe
        result = pd.DataFrame.from_dict(proportions)

        # include field of timestamps
        result['date'] = timestamps

        # return dataframe of proportion of topics for each time slice
        return result

    def get_results(self):
        """
        This method should return a table representing the evolution of each \
        topic over time.

        :return: Returns a Pandas' dataframe where each column represents a \
        timeslice and must have a `date` and columns representing each \
        topics weight in that period.
        :rtype: pd.core.frame.DataFrame
        """
        return self.results


    def get_topic_words(self, topic_id, i, n):
        """
        This method returns the top n words that better describes the \
        topic in a specific time slice.

        :param topic_id: The id of the desired topic.
        :type topic_id: int

        :param i: The position of the desired timeslice in chronological \
        order the first (oldest) time slice is indexed by 1.
        :type i: int

        :param n: This specifies how many words that better describes the \
        topic at a specific time slice should be returned.
        :type n: int

        :return: It returns a list of top n words that best describes the \
        requested topic in a specific time.
        :rtype: list[str]

        """
        return [t[0] for t in self.model.print_topic(topic_id, i - 1, n)]


    @property
    def n_timeslices(self):
        """
        This attribute should be the number of timeslices found during \
        training.

        :return: It should return the number of time slices found in corpus. \
        :rtype: int
        """
        return len(self.time_slices)


    def prepare_args(self, i):
        """
        This method should return a dictionary with all necessary values to \
        call PyLdaVis.prepare method.

        :param i: The position of the desired timeslice in chronological \
        order, the first (oldest) time slice is indexed by 1.
        :type i: int

        :return: It returns a dictionary ready to be passed to PyLdaVis
        :rtype: dict[str, any]

        """
        # Calculate parameters
        doc_topics, topic_term, doc_lengths, term_frequency, vocab = \
                self.model.dtm_vis(time=i - 1, corpus=self.bow)

        # Return dictionary with parameters
        return {
            'topic_term_dists': topic_term,
            'doc_topic_dists': doc_topics,
            'doc_lengths': doc_lengths,
            'vocab': vocab,
            'term_frequency': term_frequency,
        }


    def train(self):
        """
        Train the DTM model.

        :return: Nothing.
        :rtype: None
        """
        # prepare data to use in training
        self.bow, self.dictionary, self.timestamps, self.time_slices = \
                self.__prepare_corpus(self.corpus,
                                      self.dates,
                                      self.date_format,
                                      self.freq,
                                      self.sep)

        # train dtm model
        self.model = LdaSeqModel_(corpus=self.bow,
                                  time_slice=self.time_slices,
                                  num_topics=self.n_topics,
                                  id2word=self.dictionary)

        # calculate result
        self.results = self.__prepare_results(self.model,
                                              self.timestamps,
                                              len(self.bow),
                                              self.n_topics,
                                              self.time_slices)




from lda_over_time.lda_over_time import LdaOverTime
from lda_over_time.models.prevalence_model import PrevalenceModel

import pandas as pd

def convertScore(string: str) -> float:
    return float(string.replace(',', ''))

prefeiturasConverters = {
    'Overperforming Score (weighted  —  Likes 1x Shares 1x Comments 1x Love 1x Wow 1x Haha 1x Sad 1x Angry 1x Care 1x )':
        convertScore,
    'Sponsor Name': str,
    'Sponsor Category': str,
}

def main():
    # load prefeituras.csv
    df = pd.read_csv('~/Projects/pfg-tutorial/data/prefeituras.csv',
                     converters=prefeiturasConverters)

    df = df[['Message', 'Post Created Date']]
    df = df.dropna()

    dates = df['Post Created Date'].values
    msgs = df['Message'].values

    model = PrevalenceModel(msgs, dates, '%Y-%m-%d', '1M', 20, None, None)

    dtm = LdaOverTime(model)

    dtm.plot('')

    dtm.showvis(1)


if __name__ == '__main__':
    main()




"""
LdaOverTime is a framework that brings an easier way of doing Topic Modeling \
Analysis Over Time and get visualization of results.

In brief, Topic Modeling is a technique that finds topics that each document \
from a collection covers. And, by addind the time in this equation, we can \
study how much and why one certain topic is more or less discussed in a \
time slice.
"""
# IMPORTS
import matplotlib.pyplot as plt
import pandas as pd
import pickle
import pyLDAvis
import seaborn as sns


# TYPING
from lda_over_time.models.dtm_model_interface import DtmModelInterface
from typing import List, Optional



# CODE
class LdaOverTime:
    """
    LdaOverTime provides an easier way of taking a pre-processed set of \
    documents, choose a DTM model and get an analysis of the topic's evolution \
    over time.

    Choose a model to work with, create an instance of it by passing the \
    right parameters and then you can instantiate LdaOverTime by passing the \
    previous object.

    :param model: instance of the chosen model.
    :type model: DtmModelInterface

    :return: Nothing
    :rtype: None
    """

    def __init__(self, model: DtmModelInterface) -> None:
        """
        Initialize values and train model.
        """

        # Save parameters
        self.model = model
        self.corpus = self.model.corpus
        self.dates = self.model.dates
        self.dafe_format = self.model.date_format
        self.freq = self.model.freq
        self.n_topics = self.model.n_topics
        self.sep = self.model.sep
        self.workers = self.model.workers

        # Train model
        self.model.train()

        # Get number of time slices
        self.n_timeslices = self.model.n_timeslices

        # Get results from model
        self.results, self.dates, self.weights = self.__get_results()

        # It holds the default name of each topic
        self.topics_names = []

        # Create default topic's name: top 10 words of last time slice
        for topic in range(self.n_topics):
            words = ', '.join(
                    self.model.get_topic_words(
                        topic,
                        self.n_timeslices - 1,
                        10
                    )
            )
            self.topics_names.append(words)


    def __get_results(self):
        # extract results that will be used to plot model
        results = self.model.get_results()

        # get weights of each topic over time
        weights = results[list(range(self.n_topics))]

        # get dates
        dates = results['date'].dt.date.values

        # return dates and weights
        return results, dates, weights


    def plot(self,
             title: str,
             legend_title: Optional[str] = None,
             topic_names: Optional[List[str]] = None,
             path_to_save: Optional[str] = None,
             display: bool = True):
        """
        Plot the evolution of topics over time.

        :param title: title of plot
        :type title: str

        :param legend_title: legend's title
        :type legend_title: str, optional

        :param topic_names: custom names for each topic. If not provided, it \
        will be the top 10 words for each topic.
        :type topic_names: list[str], optional

        :param path_to_save: set it with path to save the graph. Default \
        behaviour does not save the graph.
        :type path_to_save: str, optional

        :param display: set it to False to not display graph. Default \
        behaviour is to display.
        :type display: bool, optional

        :return: Nothing
        :rtype: None
        """
        # Custom topic names were not provided: set default value (top 10 words)
        if topic_names is None:
            topic_names = self.topics_names

        # TODO: find a way of setting label names without breaking markers
        # Topic names has the right length: set new column names
        if self.weights.shape[1] == len(topic_names):
            self.weights.columns = topic_names

        # Plot
        g = sns.lineplot(data=self.weights)
        g.set_xticks(range(len(self.dates)))
        g.set_xticklabels(labels=self.dates, rotation=30)

        plt.title(title)

        plt.legend(title=legend_title,
                   bbox_to_anchor=(1,1),
                   loc="upper left")

        # Path was given: save plot in path
        if isinstance(path_to_save, str):
            plt.savefig(path_to_save)

        # Set to display: display plot
        if display is True:
            plt.show()


    def save(self, file_path: str) -> None:
        """
        Save your current work in the location file_path. You can reload your \
        work later by calling load with the same file_path.

        :param file_path: Location to save your current work.
        :type file_path: str

        :return: Nothing
        :rtype: None
        """
        with open(file_path, 'wb') as f:
            pickle.dump(self, f)


    @classmethod
    def load(cls, file_path: str) -> 'LdaOverTime':
        """
        Load your last work.

        :param file_path: Location where you saved your last work.
        :type file_path: str

        :return: Last saved work.
        :rtype: LdaOverTime
        """
        with open(file_path, 'rb') as f:
            return pickle.load(f)


    def showvis(self, time_id: int):
        """
        Show the PyLdaVis analysis of your model in a specific time slice. \
        It is useful to evaluate how good is your model.

        :param time_id: Position of the time slice from 1 to n_timeslices in \
        chronological order
        :type time_id: int

        :return: Nothing
        :rtype: None
        """
        args = self.model.prepare_args(time_id)

        display = pyLDAvis.prepare(**args)

        return pyLDAvis.display(display)


    def get_topic_words(
        self,
        topic_id: int,
        timeslice: int,
        n: int = 10
    ) -> List[str]:
        """
        Get the top `n` words of from a specific topic in the chosen \
        timeslice.

        :param topic_id: The id of the desired topic.
        :type topic_id: int

        :param timeslice: The position of the desired timeslice in \
        chronological order the first (oldest) time slice is indexed by 1.
        :type timeslice: int

        :param n: This specifies how many words that better describes the \
        topic at a specific time slice should be returned.
        :type n: int

        :return: It returns a list of top n words that best describes the \
        requested topic in a specific time.
        :rtype: list[str]
        """
        return self.model.get_topic_words(topic_id - 1, timeslice, n)


    def get_results(self) -> pd.DataFrame:
        """
        Get the model's result in format of a table.

        In this table, rows represents each time slice.

        For the columns, the `date` column holds the time slices' timestamps \
        and the remaing `n_topics` columns indexed from 1 to `n_topics` \
        holds the proportion of each topic of each time slice.

        You can get each topic's main words by calling `get_topic_words`, \
        e.g. if you want the top 10 words from the topic 3 of this table in \
        the first row, call `get_topic_words(topic_id=3, timeslice=1, n=10)`

        :return: table with results
        :rtype: pandas.DataFrame

        """
        # Get results
        results = self.model.get_results()

        # Change columns to number topics from 1 to n_topics
        results.columns = list(range(1, self.n_topics + 1)) + ['date']

        # Return table
        return results




"""
PrevalenceModel is a simpler and faster temporal LDA that returns the \
proportion of main topics in each time slice.

Its main advantage over other models is that it is fast. But it may not handle \
well the variation of the way that a topic is presented (when vocabulary to \
describe the topic varies over the given dataset).
"""
# IMPORTS
from gensim.corpora.dictionary import Dictionary
from gensim.models.ldamulticore import LdaMulticore
from lda_over_time.models.dtm_model_interface import DtmModelInterface
from multiprocessing import cpu_count
from pyLDAvis.gensim_models import _extract_data as extract_data

import pandas as pd


# TYPING
from typing import List, Optional


class PrevalenceModel(DtmModelInterface):
    """
    PrevalenceModel is a simple temporal LDA model, it is faster, but \
    it may not handle well the evolution of topics (because the vocabulary \
    used in a certain topic may vary over time).

    :param corpus:Each item from the list is one document from corpus.
    :type corpus: list[str]

    :param dates: List of timestamps for each document in corpus, each date's \
    position should match with its respective text.
    :type dates: list[str]

    :param date_format: The date format used in `dates`, e.g. "%Y/%m/%d" for \
    "YYYY/MM/DD" format. More info at `documentation`_.
    :type date_format: str

    :param freq: The frequency used to group texts, e.g. "1M15D" for a \
    frequency of a month and 15 days. Useful notations: \
    day = "D"\
    month = "M"; \
    year = "Y". \
    More info at `pandas`_
    :type freq: str

    :param n_topics: Number of topics that the DTM model should find. The \
    default value is 100.
    :type n_topics: int, optional

    :param sep: Separator used to split each word, the default value is any \
    blank space.
    :type sep: str, optional

    :param workers: Number of workers (cpus) to use. If not provided, it will \
    use the total number of threads on running machine.
    :type workers: int, optional

    :return: Nothing
    :rtype: None

    .. _documentation: https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
    .. _pandas: https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases

    """

    def __init__(self,
                 corpus: List[str],
                 dates: List[str],
                 date_format: str,
                 freq: str,
                 n_topics: int = 100,
                 sep: Optional[str] = None,
                 workers: Optional[int] = None):
        """
        I initialize the variables to train model.
        """
        # Save arguments inside object
        self.corpus = corpus
        self.dates = dates
        self.date_format = date_format
        self.freq = freq
        self.n_topics = n_topics
        self.sep = sep

        # get number of parallel workers
        self.workers = workers if isinstance(workers, int) else cpu_count()


    def __normalize_lda_model(self, corpus, lda_model):
        """
        This method creates the dataframe from the result of lda training and \
        normalizes the weights (sum of weights for a topic = 1.0)

        :param corpus: List of Gensim's BoWs, where each BoW represents one \
        text from corpus.
        :type corpus: list[BoW]

        :param lda_model: Trained lda model
        :type lda_model: gensim.models.ldamulticore.LdaMulticore

        :return: Table of normalized results of Lda model.
        :rtype: pd.core.frame.DataFrame
        """
        # Create stream of weights for each document in corpus
        stream = lda_model.get_document_topics(corpus)

        # Get weights
        lda_weights = pd.DataFrame(map(dict, stream))

        # Normalize weights
        return (lda_weights.T / lda_weights.sum(axis=1)).T


    def __prepare_corpus(self, corpus, sep):
        """
        This method takes a list of document's string and convert them to \
        correct object typing asked by Gensim's LdaMulticore.

        :param corpus: List of documents' texts
        :type corpus: list[str]

        :param sep: Character that separates each word. If not given, all \
        whitespaces are considered separators.
        :type sep: str, optional

        :return: Pair of Gensim's Dictionary instance and list of list of \
        Gensim's BoW from corpus
        :rtype: tuple[gensim.corpora.Dictionary, list[BoW]]
        """
        # Split words for each document
        data = [doc.split(sep) for doc in corpus]

        # Train Dictionary
        dictionary = Dictionary(data)

        # Convert corpus to list of BoWs
        corpus_bow = [dictionary.doc2bow(doc) for doc in data]

        # Return dictionary and list of BoWs
        return dictionary, corpus_bow


    def __prepare_model_to_prevalence(self,
                                      dates,
                                      date_format,
                                      normalized_model,
                                      n_topics):
        """
        This method prepares the model to make a prevalence analysis over the \
        lda model. For this, it will create a table with two columns: one for \
        the date and other for the main topic of each document.

        :param dates: List of timestamps for each document in corpus, each \
        date's position should match with its respective text.
        :type dates: list[str]

        :param date_format:The date format used in `dates`.
        :type date_format: str

        :param normalized_model: Normalized result from lda model over corpus.
        :type: pd.core.frame.DataFrame

        :return: Table with date and main topic columns and each row \
        represents a document from corpus.
        :rtype: pd.core.frame.DataFrame
        """
        # load dates with pandas with given format
        dates = pd.to_datetime(dates, format=date_format)

        # get the main topic for each document
        main_topics = normalized_model[list(range(n_topics))].idxmax(axis=1)

        # join dates to main topics
        zipped = list(zip(dates.values, main_topics))

        # create table with date and topic
        return pd.DataFrame(zipped, columns=['date', 'main_topic'])


    def __prepare_pyldavis(self, dates, date_format, freq, ):
        """
        This method creates the list of periods found in corpus.

        :param dates: List of timestamps for each document in corpus, each date's \
        position should match with its respective text.
        :type dates: list[str]

        :param date_format:The date format used in `dates`.
        :type date_format: str

        :param freq:The frequency used to group texts.
        :type freq: str

        :return: translation of id to date and corpus grouped by timestamp
        """
        # TODO: improve this part...
        # load dates with pandas with given format
        timestamps = pd.to_datetime(dates, format=date_format)

        # create dataframe of dates
        timestamps_df = pd.DataFrame(timestamps, columns=['date'])

        # group dates by timeslice of lenght period
        dates_by_period = timestamps_df.groupby(
                pd.Grouper(key='date', freq=freq),
                sort=True
        )

        # get timestamps used to group
        timestamp_keys = sorted(dates_by_period.groups.keys())

        # initialize variables that hold translation of id to period and
        # group corpus by period
        id_to_period = {}
        corpus_by_period = {}

        # for each period: enumerate it and extract docs that belongs to this period
        for _id, key in enumerate(timestamp_keys):

            # enumerate pediod
            id_to_period[_id + 1] = key

            # get documents that belongs to this period
            indexes = dates_by_period.get_group(key).index

            # track documents of this period
            corpus_by_period[key] = list(map(
                lambda idx: self.bow[idx],
                indexes
            ))

        # return values
        return id_to_period, corpus_by_period


    def __train_lda_model(self, corpus, dictionary, n_topics):
        """
        This method trains the lda model.

        :param corpus: List of Gensim's BoWs, where each BoW represents one \
        text from corpus.
        :type corpus: list[BoW]

        :param dictionary: Dictionary with convertion id <-> word.
        :type dictionary: gensim.corpora.dictionary.Dictionary

        :param n_topics: Number of topics that the DTM model should find. \
        The default value is 100.
        :type n_topics: int, optional

        :return: It returns the trained Gensim's model LdaMulticore.
        :rtype: gensim.models.ldamulticore.LdaMulticore
        """
        return LdaMulticore(corpus=corpus,
                            id2word=dictionary,
                            num_topics=n_topics,
                            random_state=100,
                            chunksize=100,
                            passes=10,
                            minimum_probability=0.0,
                            per_word_topics=True,
                            workers=self.workers)


    def __train_prevalence(self,
                           docs_date_main_topic,
                           freq,
                           grouped_by_period):
        """
        This method takes the result of LDA and calculates the model of \
        temporal LDA with prevalence (frequency of main topics in a period).

        :param docs_date_main_topic: Table of documents with their date and \
        main topic
        :type docs_date_main_topic: pd.core.frame.DataFrame

        :param freq: Frequency to group documents
        :type freq: str

        :param grouped_by_period: Documents grouped by frequency
        :type grouped_by_period: pd.core.groupby.DataFrameGroupBy

        :return: It returns the temporal LDA trained model.
        :rtype: pd.core.frame.DataFrame
        """
        # Calculate number of posts per period
        posts_per_period = grouped_by_period.size().reset_index(name='count')

        # Group by period and topic
        grouped_by_period_topic = docs_date_main_topic.groupby(
            [
                pd.Grouper(key='date', freq=freq),
                pd.Grouper(key='main_topic')
            ],
            sort=True
        )

        # Calculate frequency of each main topic in a time interval period
        frequency = grouped_by_period_topic. \
                        size(). \
                        reset_index(name='count'). \
                        pivot(index='date',
                              columns='main_topic',
                              values='count').reset_index()

        # Replace nan with zero
        frequency = frequency.fillna(0.0)

        # Create table of prevalence
        prevalence = pd.DataFrame(
                (frequency[list(range(self.n_topics))].values.T / \
                                      posts_per_period['count'].values).T
        )

        # Add column with dates
        prevalence['date'] = frequency['date'].values

        # Return temporal lda model
        return prevalence


    @property
    def n_timeslices(self):
        """
        This attribute should be the number of timeslices found during \
        training.

        :return: It should return the number of time slices found in corpus. \
        :rtype: int
        """
        return self.grouped_by_frequency.ngroups


    def get_results(self):
        """
        This method should return a table representing the evolution of each \
        topic over time.

        :return: Returns a Pandas' dataframe where each column represents a \
        timeslice and must have a `date` and columns representing each \
        topics weight in that period.
        :rtype: pd.core.frame.DataFrame
        """
        return self.temporal_lda_model


    def get_topic_words(self, topic_id, i, n):
        """
        This method should return the top n words that better describes the \
        topic in a specific time slice.

        :param topic_id: The id of the desired topic.
        :type topic_id: int

        :param i: The position of the desired timeslice in chronological \
        order, the first (oldest) time slice is indexed by 1.
        :type i: int

        :param n:This specifies how many words that better describes the topic \
        at a specific time slice should be returned.
        :type n: int

        :return: It returns a list of top n words that best describes the \
        requested topic in a specific time.
        :rtype: list[str]
        """
        return [
                self.dictionary[word[0]]
                for word in self.lda_model.get_topic_terms(
                        topic_id,
                        n
                )
        ]


    def prepare_args(self, i):
        """
        This method should return a dictionary with all necessary values to \
        call PyLdaVis.prepare method.

        :param i: The position of the desired timeslice in chronological \
        order, the first (oldest) time slice is indexed by 1.
        :type i: int

        :return: It returns a dictionary ready to be passed to PyLdaVis
        :rtype: dict[str, any]
        """
        # get timestamp
        period = self.id_to_timestamp[i]

        # return prepared data
        return extract_data(self.lda_model,
                            self.timestamp_to_corpus[period],
                            self.dictionary)


    def train(self):
        """
        This method trains the dtm model.

        :return: Nothing.
        :rtype: None
        """
        # Get dictionary and convert corpus to correct type
        self.dictionary, self.bow = self.__prepare_corpus(self.corpus,
                                                          self.sep)

        # train lda model
        self.lda_model = self.__train_lda_model(self.bow,
                                                self.dictionary,
                                                self.n_topics)

        # get normalized results of lda model from corpus
        self.normalized = self.__normalize_lda_model(self.bow,
                                                     self.lda_model)

        # prepare model to prevalence analysis
        self.docs_date_main_topic = \
                self.__prepare_model_to_prevalence(self.dates,
                                                   self.date_format,
                                                   self.normalized,
                                                   self.n_topics)
        # group documents by date
        self.grouped_by_frequency = \
                self.docs_date_main_topic.groupby(pd.Grouper(key='date',
                                                             freq=self.freq),
                                                  sort=True)

        # calculate temporal lda model
        self.temporal_lda_model = \
                self.__train_prevalence(self.docs_date_main_topic,
                                        self.freq,
                                        self.grouped_by_frequency)

        # get the list of timestamps
        self.id_to_timestamp, self.timestamp_to_corpus = \
                self.__prepare_pyldavis(self.dates, self.date_format, self.freq)

